#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "ppport.h"
#if defined(HEIMDAL)
#include <gssapi.h>
#endif

#if !defined(HEIMDAL)
#include <gssapi/gssapi.h>
#endif






#define _DBSIZE 256
#define _DBG( message )if ( debuggingmode ) { _debug_message( message ); }

typedef gss_buffer_desc *p_namebuffer;
typedef gss_ctx_id_t *p_gss_ctx_id_t;

static gss_OID_desc krb5_oid_desc = {9, (void *) "\x2a\x86\x48\x86\xf7\x12\x01\x02\x02"};

static int debuggingmode = 0;

void _debug_message( char *s) {
   printf("\nAuthen::SPNEGO::GSSAPI  ");
   printf( s );
}


MODULE = Authen::SPNEGO::GSSAPI		PACKAGE = Authen::SPNEGO::GSSAPI
PROTOTYPES: ENABLE
###############################################################################

INCLUDE: headerfilegeneratedconstantwrappers.xsh

###############################################################################
MODULE = Authen::SPNEGO::GSSAPI		PACKAGE = Authen::SPNEGO::GSSAPI
void gss_display_status_GSS_CODE( in_status_value )
                int in_status_value
    INIT:
       gss_buffer_desc status_string;
       OM_uint32 minor_status;
       OM_uint32 major_status;
       int message_context = 0;
    PPCODE:
       status_string.length = _DBSIZE;
       status_string.value = alloca(  _DBSIZE );
       do {
          major_status = gss_display_status( &minor_status,
                                             in_status_value,
                                             GSS_C_GSS_CODE,
					     GSS_C_NULL_OID,
					     &message_context,
					     &status_string );
          if ( major_status == GSS_S_COMPLETE ) {
               XPUSHs( sv_2mortal( newSVpv( status_string.value, 0  ) )  );
          }
          if ( major_status == GSS_S_BAD_MECH) {
               XPUSHs( sv_2mortal( newSVpv( "GSS_S_BAD_MECH returned from gss_display_status", 0  ) )  );
          }
          if ( major_status == GSS_S_BAD_STATUS) {
               XPUSHs( sv_2mortal( newSVpv( "GSS_S_BAD_STATUS status value was not recognized", 0  ) )  );
          }
	} while (    message_context != 0
	          && major_status == GSS_S_COMPLETE);

###############################################################################
MODULE = Authen::SPNEGO::GSSAPI		PACKAGE = Authen::SPNEGO::GSSAPI
void gss_display_status_MECH_CODE( in_status_value )
                int in_status_value
    INIT:
       gss_buffer_desc status_string;
       OM_uint32 minor_status;
       OM_uint32 major_status;
       int message_context = 0;
    PPCODE:
       status_string.length = _DBSIZE;
       status_string.value = alloca(  _DBSIZE );
       _DBG( "gss_display_status_MECH_CODE" )
       do {
          major_status = gss_display_status( &minor_status,
                                             in_status_value,
                                             GSS_C_MECH_CODE,
					     GSS_C_NULL_OID,
					     &message_context,
					     &status_string );
          if ( major_status == GSS_S_COMPLETE ) {
               XPUSHs( sv_2mortal( newSVpv( status_string.value, 0  ) )  );
          }
          if ( major_status == GSS_S_BAD_MECH) {
               XPUSHs( sv_2mortal( newSVpv( "GSS_S_BAD_MECH returned from gss_display_status", 0  ) )  );
          }
          if ( major_status == GSS_S_BAD_STATUS) {
               XPUSHs( sv_2mortal( newSVpv( "GSS_S_BAD_STATUS status value was not recognized", 0  ) )  );
          }
	} while (    message_context != 0
	          && major_status == GSS_S_COMPLETE);

###############################################################################
MODULE = Authen::SPNEGO::GSSAPI		PACKAGE = Authen::SPNEGO::GSSAPI
void achim_gss_init_sec_context( server_gss_name, in_sec_context)
                    gss_name_t     server_gss_name
		    p_gss_ctx_id_t in_sec_context


     INIT:
        OM_uint32 major_status;
        OM_uint32 minor_status;
        gss_ctx_id_t context_handle;
        gss_buffer_desc *token_ptr;
        gss_buffer_desc send_token;
        int flags;
        p_gss_ctx_id_t p_out_context_handle;
	gss_buffer_t return_token;
    PPCODE:
        context_handle =  *in_sec_context;
        token_ptr = GSS_C_NO_BUFFER;
        major_status = gss_init_sec_context( &minor_status,
                                             GSS_C_NO_CREDENTIAL,
                                             &context_handle,
                                             server_gss_name,
                                             &krb5_oid_desc ,
                                             GSS_C_REPLAY_FLAG,
                                             GSS_C_INDEFINITE,
					     GSS_C_NO_CHANNEL_BINDINGS,
					     GSS_C_NO_BUFFER,
                                             NULL, /* actual mech type */
					     &send_token,
                                             &flags,
					     NULL);
        if ( debuggingmode ) {
	   char *b;
	   b = alloca( _DBSIZE );

           _debug_message("achim_gss_init_sec_context( server_gss_name )");
	   if ( snprintf( b, _DBSIZE, "major_status %x", major_status) > 0 ) {
	       _debug_message( b );
	   } else {
	       Perl_croak(aTHX_ "snprintf failed in Authen::SPNEGO::GSSAPI::achim_gss_init_sec_context");
	   }
	   if ( snprintf( b, _DBSIZE, "minor_status %x", minor_status) > 0 ) {
	       _debug_message( b );
	   } else {
	       Perl_croak(aTHX_ "snprintf failed in Authen::SPNEGO::GSSAPI::achim_gss_init_sec_context");
	   }
        }
        XPUSHs( sv_2mortal( newSVnv( major_status ) )  );
        XPUSHs( sv_2mortal( newSVnv( minor_status ) )  );
	if ( major_status == GSS_S_COMPLETE ) {
	      p_out_context_handle = malloc( sizeof( gss_ctx_id_t ) );
	      if (  p_out_context_handle == NULL ) {
	         Perl_croak(aTHX_ "malloc 1");
	      }
	      *p_out_context_handle = context_handle;
              XPUSHs( sv_newmortal());
              sv_setref_pv(ST(2), "p_gss_ctx_id_t", (void*)p_out_context_handle);
              /* ---------------------- */
	      return_token = malloc( sizeof( gss_buffer_desc ));
	      if ( return_token == NULL ) {
	           Perl_croak(aTHX_ "malloc 2");
	      }
	      return_token->length = send_token.length;
	      return_token->value = send_token.value;
	      XPUSHs( sv_newmortal());
	      sv_setref_pv(ST(3), "gss_buffer_t", (void*)return_token);
	}

#-------------------------------------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI		PACKAGE = Authen::SPNEGO::GSSAPI
void major_status_fields( major_status )
                 OM_uint32 major_status
        PPCODE:
	     XPUSHs( sv_2mortal( newSVnv( GSS_CALLING_ERROR(major_status)  ) )  );
	     XPUSHs( sv_2mortal( newSVnv( GSS_ROUTINE_ERROR(major_status)  ) )  );
	     XPUSHs( sv_2mortal( newSVnv( GSS_SUPPLEMENTARY_INFO(major_status)  ) )  );

#-------------------------------------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = gss_name_t PREFIX=authen_spnego_gssapi_

void authen_spnego_gssapi_DESTROY( servername )
         gss_name_t servername
   CODE:
      OM_uint32 major_status;
      OM_uint32 minor_status;
      char *b1;
     if ( debuggingmode ) {
          _debug_message("authen_spnego_gssapi_DESTROY( gss_name_t )");
      }
      major_status = gss_release_name( &minor_status,
                                       &servername );
      if ( major_status != GSS_S_COMPLETE ) {
          b1 = alloca(_DBSIZE);
          snprintf( b1, _DBSIZE,
                    "Fatal Error while Calling GSSAPI  gss_release_name major %u minor %u",
                     major_status, minor_status );
          Perl_croak(aTHX_ b1);
      }

#-----------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = Authen::SPNEGO::GSSAPI
void  gss_context_time( p_in_context )
                       p_gss_ctx_id_t p_in_context

     PPCODE:
        OM_uint32 major_status;
        OM_uint32 minor_status;
        OM_uint32 time_rec;
        major_status = gss_context_time( &minor_status,
	                                 *p_in_context,
					 &time_rec );
        XPUSHs( sv_2mortal( newSVuv( major_status ) )  );
        XPUSHs( sv_2mortal( newSVuv( minor_status ) )  );
	if ( major_status == GSS_S_COMPLETE ) {
	    XPUSHs( sv_2mortal( newSVuv( time_rec) )  );
	}



#-----------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = Authen::SPNEGO::GSSAPI PREFIX=authen_spnego_gssapi_
void authen_spnego_gssapi_gss_import_name( in_servername )
                 char * in_servername

     INIT:
        OM_uint32 major_status;
        OM_uint32 minor_status;
        gss_name_t server_gss_name;
	gss_buffer_desc snbuffer;
      PPCODE:
        snbuffer.value = strdup( in_servername );
	if (  snbuffer.value == NULL ) {
	   Perl_croak(aTHX_ "strdup");
	}
        snbuffer.length = strlen( snbuffer.value ) ;
        major_status = gss_import_name( &major_status,
                                        &snbuffer,
                                        GSS_C_NULL_OID,
                                        &server_gss_name);
        XPUSHs( sv_2mortal( newSVuv( major_status ) )  );
        XPUSHs( sv_2mortal( newSVuv( minor_status ) )  );
	free( snbuffer.value );
        if ( major_status == GSS_S_COMPLETE) {
           XPUSHs( sv_newmortal());
           sv_setref_pv(ST(2), "gss_name_t", (void*)server_gss_name);
        }


#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = Authen::SPNEGO::GSSAPI
size_t  gss_buffer_desc_length( p_gss_buffer )
                 gss_buffer_t p_gss_buffer
        CODE:
	     RETVAL = p_gss_buffer->length;
        OUTPUT:
	     RETVAL


#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = Authen::SPNEGO::GSSAPI
size_t  __dump_token2stream( stream, gssapibuffer )
             FILE * stream
	     gss_buffer_t gssapibuffer
       CODE:
           RETVAL = fwrite( gssapibuffer->value,
	                    1,
	                    gssapibuffer->length,
			    stream);
       OUTPUT:
           RETVAL

#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = p_gss_ctx_id_t PREFIX=authen_spnego_gssapi_
void authen_spnego_gssapi_DESTROY( p_scurity_context )
                 p_gss_ctx_id_t p_scurity_context
	CODE:
	    OM_uint32 major_status;
            OM_uint32 minor_status;
	    gss_buffer_t output_token;

	    if ( debuggingmode ) { _debug_message("authen_spnego_gssapi_DESTROY(p_scurity_context  )"); }
	    if ( *p_scurity_context !=GSS_C_NO_CONTEXT ) {
		  _DBG("  ah, not a GSS_C_NO_CONTEXT" )
		  output_token = GSS_C_NO_BUFFER;
		  major_status = gss_delete_sec_context( &minor_status,
		                                         p_scurity_context,
							 output_token );
                  if ( major_status != GSS_S_COMPLETE ) {
		       Perl_croak(aTHX_ "gss_delete_sec_context failed");
		  } else {
		     _DBG( "  succesfully deleted seccontext")
		     if ( output_token != GSS_C_NO_BUFFER ) {
		        _DBG( "  throwing away Close token..." )
			major_status = gss_release_buffer( &minor_status, output_token );
                        if ( major_status != GSS_S_COMPLETE ) {
		           Perl_croak(aTHX_ "gss_delete_sec_context failed");
		        }
		     }
		  }
	    }
	    free( p_scurity_context );

#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = gss_buffer_t PREFIX=authen_spnego_gssapi_
void
authen_spnego_gssapi_DESTROY( gsaapibuffer )
                 gss_buffer_t gsaapibuffer
	CODE:
	    OM_uint32 major_status;
            OM_uint32 minor_status;
            _DBG(  "authen_spnego_gssapi_DESTROY(gsaapibuffer )" )
	    if ( gsaapibuffer->length > 0 ) {
	           major_status = gss_release_buffer( &minor_status, gsaapibuffer );
                   if ( major_status != GSS_S_COMPLETE ) {
		          Perl_croak(aTHX_ "gss_release_buffer failed");
		   } else {
		      _DBG( " gss_delete_sec_context OK")
		   }
	    }
	    free( gsaapibuffer );



#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = Authen::SPNEGO::GSSAPI
p_gss_ctx_id_t GSS_C_NO_CONTEXT()

     CODE:
         p_gss_ctx_id_t r;
         r = malloc( sizeof( gss_ctx_id_t ) );
	 if ( r == NULL ) { Perl_croak(aTHX_ "malloc"); }
	 *r = GSS_C_NO_CONTEXT;
	 RETVAL = r;
     OUTPUT:
         RETVAL


#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = Authen::SPNEGO::GSSAPI
void
array_gss_indicate_mechs()
    INIT:
        OM_uint32 major_status;
        OM_uint32 minor_status;
	gss_OID_set mechset;
	gss_OID tmp_oid;
	unsigned int i;
	unsigned int n;
	unsigned char *oidelem;
	AV *tmparray;
     PPCODE:
        _DBG("going to call gss_indicate_mechs...");
        major_status =  gss_indicate_mechs( &minor_status, &mechset);
	if (major_status) {
	   _DBG("failed!");
	} else {
	   _DBG("OK");
	   for (i=0; i < mechset->count; i++) {
	       _DBG("found");
	       tmparray = newAV() ;
	       tmp_oid = &mechset->elements[i];
	       oidelem = tmp_oid->elements;
               for ( n=0; n < tmp_oid->length; n++) {
		   av_push(  tmparray, newSVuv( *oidelem )   );
		   oidelem++;
	       }
	    XPUSHs( sv_2mortal(  newRV((SV *) tmparray ) )   ) ;

	   }

	}


#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = Authen::SPNEGO::GSSAPI
void enable_internal_debuggig()
      CODE:
          debuggingmode = 1;
	  _DBG( "Debugging output enabeled");

#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = Authen::SPNEGO::GSSAPI
void disable_internal_debuggig()
      CODE:
          _DBG( "Disabeling debugging output");
          debuggingmode = 0;

#-------------------------------------------------
MODULE = Authen::SPNEGO::GSSAPI PACKAGE = p_namebuffer PREFIX=authen_spnego_gssapi_

void authen_spnego_gssapi_DESTROY( gssbuffer )
         p_namebuffer gssbuffer
   CODE:
      _DBG( "authen_spnego_gssapi_DESTROY( p_namebuffer )");
      free( gssbuffer->value);
      free( gssbuffer );
