use 5.008004;
use ExtUtils::MakeMaker;
# See lib/ExtUtils/MakeMaker.pm for details of how to influence
# the contents of the Makefile that is written.

use strict;

my @whichcand = qw(/bin/which /usr/bin/which);

my $CONSTINCLUDE = 'headerfilegeneratedconstantwrappers.xsh';
my $MAINPACKAGENAME = 'Authen::SPNEGO::GSSAPI';

my @SEARCH_HEADER_DIR = map { ($_, $_ . '/gssapi') } qw(/usr/include /usr/include/local);

use Getopt::Long;
use vars qw($opt_gssapiimpl $opt_help);

&GetOptions("gssapiimpl=s", \$opt_gssapiimpl, "help", \$opt_help);


if ($opt_help){
print <<HELP;

==================================================================
BUILDING HELP
==================================================================

Hello - you obviously need Help! :-)

This additional command line options are available
for adjusting:

(1) --gssapiimpl /bla/local/heimdal
  This tells me where to look for your installed
  GSSAPI installation MIT or Heimdal.
  This must be the directory that contains
  the ./lib and ./include directories

HELP
;
  exit;
}


#-------------------------------------------------------------
print "\n Welcome to Authen::SPNEGO::GSSAPI setup! \n";

my $ADDLINKPATH='';
my $ADDINCLUDEPATH='';

if ($opt_gssapiimpl) {
    print "\nTrying GSSAPI dir $opt_gssapiimpl...";
    checkdir ( $opt_gssapiimpl );
    #
    my $libdir = join '/', $opt_gssapiimpl, 'lib';
    checkdir( $libdir );
    #
    my $incdir = join '/', $opt_gssapiimpl, 'include';
    -e $incdir || fatal_error( "$incdir does not exist" );
    unless ( -e join '/', $incdir, 'gssapi.h' ) {
        $incdir = join '/', $incdir, 'gssapi';
    }
    checkdir( $incdir );
    $ADDLINKPATH="-L$libdir";
    $ADDINCLUDEPATH=$incdir;
    print "OK \n  using GSSAPI libs in $ADDLINKPATH \n  using GSSAPI includedir $ADDINCLUDEPATH\n";
} else {
    foreach ( @SEARCH_HEADER_DIR ) {
       print("\n  trying $_ ...");
       test_if_dir_available( $_ );
       my $fgssapih = join '/', $_, 'gssapi.h';
       if ( -e $fgssapih && -f $fgssapih && -r $fgssapih  ) {
          printf " found $fgssapih";
	  $ADDINCLUDEPATH = $_;
	  last;
       }
    }
    $ADDINCLUDEPATH || die "cannot find gssapi.h!";
}
#------------------------------------------------------
#------------------------------------------------------
my $FILEGSSAPI_H = join '/', $ADDINCLUDEPATH, 'gssapi.h';
-e $FILEGSSAPI_H || fatal_error( "$FILEGSSAPI_H does not exist" );
-f $FILEGSSAPI_H || fatal_error( "$FILEGSSAPI_H is dir, not plain file" );
-r $FILEGSSAPI_H || fatal_error( "$FILEGSSAPI_H is not readable" );
print "\n  using GSSAPI headers from $FILEGSSAPI_H \n";

#---------------------------------------------------------------
my @GSSLIBS = ();

#---------------------------------------------------------------

print "\n Searching krb5-config command...";
my $krb5cmd = find_krb5config_cmd($opt_gssapiimpl );
unless ($krb5cmd) {
    fatal_error( 'not found!');
    exit
} else {
  print $krb5cmd ;
}
#---------------------------------------------
my $libconf =  krb5_libconfig( $krb5cmd );
unless ($libconf) {
   fatal_error("$krb5cmd does not respond libconf!");
} else {
  push @GSSLIBS, $libconf ;

}
print "\n Using libconfiguration", @GSSLIBS;
#---------------------------------------------------------------
my $incconf =  krb5_cflags( $krb5cmd  );
unless ( $incconf ) {
   fatal_error("$krb5cmd does not respond inconfconf!");
}
print "\n Using incconfiguration", $incconf;
#---------------------------------------------------------------

my @EXTRADEFINES;
if ( is_heimdal($krb5cmd) ) {
   push @EXTRADEFINES, '-DHEIMDAL';
}
print "\n using own defines ", @EXTRADEFINES;
#---------------------------------------------------------------


my @GSS_S = cannibalize_header($FILEGSSAPI_H);
print "\n Importing constants ", join ' ', @GSS_S;
print "\n Writing $CONSTINCLUDE...";
open XSHFILE, ">$CONSTINCLUDE" or die $!;
print XSHFILE join "\n", map { gen_constwrapfunction('OM_uint32',
                            $_,
                            $MAINPACKAGENAME,
			    $MAINPACKAGENAME) } @GSS_S;
close XSHFILE;
print 'OK';

print "\n Writing GSSSExports...";
open PMFILE, '>lib/Authen/SPNEGO/GSSAPI/GSSSExports.pm' or die $!;
print PMFILE join "\n",
             "package $MAINPACKAGENAME" . '::GSSSExports;',
	     "use strict;",
	     "sub gss_s_constant_exports {",
	     'return qw(',  join ' ', @GSS_S  ,  ');',
	     "}",
	     "1;";
close  PMFILE;


#------------------------------------------------------
sub checkdir {
  my ($dir) = @_;
    -e $dir || fatal_error( "$dir does not exist" );
    -d $dir || fatal_error( "$dir not directory" );
    -r $dir || fatal_error( "$dir not readable" );
}
#------------------------------------------------------
sub fatal_error {
    print 'Fatal Error:', @_, "\n";
    exit;
}
#------------------------------------------------------



WriteMakefile(
    NAME              => 'Authen::SPNEGO::GSSAPI',
    VERSION_FROM      => 'lib/Authen/SPNEGO/GSSAPI.pm', # finds $VERSION
    #PREREQ_PM         => { 'MIME::Base64' => 3.01  }, # e.g., Module::Name => 1.1
    ($] >= 5.005 ?     ## Add these new keywords supported since 5.005
      (ABSTRACT_FROM  => 'lib/Authen/SPNEGO/GSSAPI.pm', # retrieve abstract from module
       AUTHOR         => 'Achim Grolms <achim@grolmsnet.de>') : ()),
       LIBS => join (' ', @GSSLIBS),
       INC  => $incconf,
       DEFINE            => join (' ', @EXTRADEFINES) # e.g., '-DHAVE_SOMETHING'

	# Un-comment this if you add C files to link with later:
    # OBJECT            => '$(O_FILES)', # link all the C files too
);



# ---------------------------------------------------------------------------
sub test_if_dir_available {
   my ($dir) = @_;
   return -e $dir && -d $dir && -r $dir;
}
# ---------------------------------------------------------------------------
sub cannibalize_header {
   my ($gssapiheader) = @_;
   $gssapiheader || die missing param '$gssapiheader';
   -e $gssapiheader || die "$gssapiheader does not exist!";
   my @r;
   open INGSSAPIH, $gssapiheader or die "error while opening $gssapiheader : $!";
   while(<INGSSAPIH>) {
       if (m/#define[ ]+(GSS_S_[A-Z_a-z]+)/) {
	  push @r, $1
       }
   }
   close INGSSAPIH or die $!;
   return @r;
}
# ---------------------------------------------------------------------------
sub gen_constwrapfunction {
   my ($returntyp, $constant, $modul, $package) = @_;
   return join "\n",,
	       "MODULE = $modul PACKAGE = $package",
	       "",
	       "$returntyp $constant()",
               "CODE:",
               "        RETVAL = $constant;",
               "OUTPUT:",
               "        RETVAL",
	       "\n";

}
# ---------------------------------------------------------------------------
#-------------------------------------------------
sub find_which_command {
    foreach (@whichcand) {
       return $_ if ( -e $_);
    }
    return undef;
}
#-------------------------------------------------
sub find_krb5config_cmd {
   my ($expl_path) = @_;
   my $r = undef;
   unless ($expl_path) {
      my $whichcmd =  find_which_command() || die 'cannot locat which command';
      $r = `$whichcmd krb5-config 2>/dev/null`;
      chomp $r;
   } else {
      my $supposed = join '/', $expl_path, 'bin', 'krb5-config';
      if ( -e $supposed ) {
         $r = $supposed ;
      }
   }
   return $r;
}
#-------------------------------------------------
sub krb5_libconfig {
   my ($cfgcmd) = @_;
   $cfgcmd || die 'no $cfgcmd ';
   return `$cfgcmd --libs gssapi`;
}
#-------------------------------------------------
sub krb5_cflags  {
   my ($cfgcmd) = @_;
   $cfgcmd || die 'no $cfgcmd ';
   return `$cfgcmd --cflags gssapi`;
}
#-------------------------------------------------
#-------------------------------------------------
sub krb5_version {
   my ($cfgcmd) = @_;
   $cfgcmd || die 'no $cfgcmd ';
   return `$cfgcmd --version`;
}
#-------------------------------------------------
sub is_heimdal {
   my ($cfgcmd) = @_;
   my $r = undef;
   my @vinfo = krb5_version( $cfgcmd );
   GEFUNDEN: {
      foreach ( @vinfo) {
         if ( m/eimdal/) {
            $r = 1;
	    last GEFUNDEN;
         }
       }
   }
   return $r;
}
#-------------------------------------------------